/*
 * File: file.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */


/***

The routines in this file handle the commands for the reading, writing
and lookup of disk files.  All of details about the reading and writing
of the disk are in "fileio.c".

***/

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	"rho.h"
#include	"file.h"
#include	"buffer.h"
#include	"display.h"
#include	"exec.h"
#include	"fileio.h"
#include	"line.h"
#include	"messages.h"
#include	"terminal.h"
#include	"utils.h"
#include	"window.h"

extern int readlink(const char *, char *, size_t);
extern int unlink(const char *);
extern int chown(const char *, uid_t, gid_t);
extern char *basename(const char *);

/*++

Interactively read a file into the current buffer.  Find the name of
the file and calls readin()

++*/
int read_file(f, n)
int f, n;

{
    char *fname;                /* file name to read */

    f = n = 0;                  /* splint */

    if ((fname = prompt4filename(TEXT131)) == NULL)     /* "Read file" */
        return (FALSE);

    return (readin(fname, TRUE));
}



/*++

Insert file "fname" into the current buffer, Called by
insert_file(). Return the final status of the read.

++*/
int insertfile(fname)
char *fname;
{
    BUFFER *bp;
    char mesg[NSTRING];
    int cmark;                  /* current mark */
    int i;
    int nbytes;
    int nline;
    int s;
    LINE *lp0;
    LINE *lp1;
    LINE *lp2;

    bp = curbp;
    bp->b_flag |= BFCHG;        /* We have changed       */
    bp->b_flag &= ~BFINVS;      /* and are not temporary */
    if ((s = ffropen(fname)) == FIOERR) /* Hard file open.       */
        goto out;
    if (s == FIOFNF) {          /* File not found.       */
        mlwrite(TEXT152);       /* "[No such file]" */
        return (FALSE);
    }
    mlwrite(TEXT153);           /*"[Inserting file]" */

    /* back up a line and save the mark here */
    curwp->w_dotp = lback(curwp->w_dotp);
    curwp->w_doto = 0;
    for (cmark = 0; cmark < NMARKS; cmark++) {
        curwp->w_markp[cmark] = curwp->w_dotp;
        curwp->w_marko[cmark] = 0;
    }

    nline = 0;
    while ((s = ffgetline(&nbytes)) == FIOSUC) {
        if ((lp1 = alloc_line((size_t) nbytes)) == NULL) {
            s = FIOMEM;
            break;
        }
        lp0 = curwp->w_dotp;    /* line previous to insert */
        lp2 = lp0->l_fp;        /* line after insert */

        /* re-link new line between lp0 and lp2 */
        lp2->l_bp = lp1;
        lp0->l_fp = lp1;
        lp1->l_bp = lp0;
        lp1->l_fp = lp2;

        /* and advance and write out the current line */
        curwp->w_dotp = lp1;
        for (i = 0; i < nbytes; ++i)
            lputc(lp1, i, fline[i]);
        ++nline;
    }
    ffclose();                  /* Ignore errors.       */
    curwp->w_markp[0] = lforw(curwp->w_markp[0]);
    strcpy(mesg, "[");
    if (s == FIOERR) {
        strcat(mesg, TEXT141);  /* "I/O ERROR, " */
        curbp->b_flag |= BFTRUNC;
    }
    if (s == FIOMEM) {
        strcat(mesg, TEXT142);  /* "OUT OF MEMORY, " */
        curbp->b_flag |= BFTRUNC;
    }
    strcat(mesg, TEXT154);      /* "Inserted " */
    strcat(mesg, int2str(nline));
    strcat(mesg, TEXT143);      /*" line" */
    if (nline > 1)
        strcat(mesg, "s");
    strcat(mesg, "]");
    mlwrite(mesg);

  out:
    /* advance to the next line and mark the window for changes */
    curwp->w_dotp = lforw(curwp->w_dotp);
    curwp->w_flag |= WFHARD | WFMODE;

    /* copy window parameters back to the buffer structure */
    curbp->b_dotp = curwp->w_dotp;
    curbp->b_doto = curwp->w_doto;
    for (cmark = 0; cmark < NMARKS; cmark++) {
        curbp->b_markp[cmark] = curwp->w_markp[cmark];
        curbp->b_marko[cmark] = curwp->w_marko[cmark];
    }
    curbp->b_fcol = curwp->w_fcol;

    if (s == FIOERR)            /* False if error.      */
        return (FALSE);
    return (TRUE);
}


/*++

Interactively insert a file into the buffer.  Get the
filename and call ifile()

++*/
int insert_file(f, n)
int f, n;

{
    int s;
    char *fname;                /* file name */
    LINE *curline;

    f = n = 0;                  /* splint */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */

    if ((fname = prompt4filename(TEXT132)) == NULL)     /* "Insert file" */
        return (FALSE);
    /*
       Save the local pointers to hold global ".", in case
       $yankflag is set to 1.  Insert-file always places the
       starting offset point at 0.  Hold *previous* line
       position, since the current line may be re-allocated.
     */
    if (yankflag)
        curline = lback(curwp->w_dotp);

    s = insertfile(fname);

    if (yankflag)
        curwp->w_dotp = lforw(curline);

    return (s);
}



/*++

Selects a file for editing.  Checks other buffers first. Else read it
in if it exists.  Else new buffer.  All work done by getfile() which
uses readin()

++*/
int find_file(f, n)
int f, n;

{
    char *fname;

    if ((fname = prompt4filename(TEXT133)) == NULL)     /* "File name" */
        return (FALSE);
    f = getfile(fname, TRUE);
    if ((n == 2) && (f)) {
        window_ro();
    }
    return f;

}


/*++

find_file wrapper putting buffer in readonly VIEW mode.

++*/
int view_file(f, n)
int f, n;

{
    n = 2;
    return (find_file(f, n));
}


/*++

Checks for filename in buffers and switches to that buffer.
Else creates new buffer and hands off the filename to readin()
Global helper.

++*/
int getfile(fname, lockfl)
char fname[];                   /* file name to find */
int lockfl;                     /* check for locks in readin() */

{
    BUFFER *bp;
    LINE *lp;
    int i;

    /* bring to top if already in existing buffer ... */
    for (bp = bheadp; bp != NULL; bp = bp->b_bufp) {
        if ((bp->b_flag & BFINVS) == 0 && strcmp(bp->b_name, fname) == 0) {
            make_bp_top(bp);
            lp = curwp->w_dotp;
            i = curwp->w_ntrows / 2;
            while (i-- && lback(lp) != curbp->b_linep)
                lp = lback(lp);
            curwp->w_linep = lp;
            curwp->w_flag |= WFMODE | WFHARD;
            mlwrite(TEXT135);   /* "[Old buffer]" */
            return (TRUE);
        }
    }

    /* ... or load into new buffer */
    bp = exist_buffer(fname, TRUE, 0);
    if (bp == NULL) {
        mlwrite(TEXT137);       /*"Cannot create buffer" */
        return (FALSE);
    }
    make_bp_top(bp);            /* switch to the new buffer */
    return (readin(fname, lockfl));     /* Read in file.        */
    return (TRUE);
}


/*++

Read file "fname" into the current buffer, blowing away any text
found there.  Called by both the read and find commands.  Return
the final status of the read.  Also called by the mainline, to
read in a file specified on the command line as an argument. 
The command in $readhook is called after the buffer is set up
and before it is read.
Local helper for readin() and writeout()

++*/
static int expandlink(fname, linkedfile)
char *fname;                    /* name of file */
char *linkedfile;               /* name of path of symbolic link */
{
    char tmplink[1024];
    struct stat st;
    int index, iter;

    iter = 0;
    strcpy(tmplink, fname);

    while (TRUE) {

        /* avoid infinite loops on cyclic symbolic links */
        if (iter >= 32) {
            return (-1);
        }

        /* get the permissions on the file */
        /* return if file does not exist (creating new file?) */
        if (lstat(tmplink, &st) != 0) {
            strcpy(linkedfile, tmplink);
            return (0);
        }
        /* also return if file is not a symbolic link */
        if (!S_ISLNK(st.st_mode)) {
            strcpy(linkedfile, tmplink);
            return ((int) st.st_mode & (int) S_IFMT);
        }
        /* otherwise iterate */
        ++iter;

        /* get name of file pointed to */
        index = readlink(tmplink, linkedfile, 1024);
        linkedfile[index] = '\0';

        if (*linkedfile != '/') {
            index = (int) strlen(tmplink);
            while (index > 0) {
                --index;
                if (tmplink[index] == '/')
                    index = 0;
                else
                    tmplink[index] = '\0';
            }
            strcat(tmplink, linkedfile);
        } else
            strcpy(tmplink, linkedfile);
    }
}



/*++

This function does the actual managing of the reading in of a file.
Sits on top of the fileio fns.

++*/
int readin(fname, lockfl)
char fname[];                   /* name of file to read */
int lockfl;                     /* check for file locks? */

{
    LINE *lp1;
    LINE *lp2;
    int i;
    WNDW *wp;
    BUFFER *bp;
    int s;
    int nline;
    int cmark;                  /* current mark */
    int nbytes;
    char mesg[NSTRING];
    char fnlink[1024];

    lockfl = 0;                 /* splint */

    s = expandlink(fname, fnlink);

    if ((s == 0) || (s == (int) S_IFREG)) {
        strcpy(fname, fnlink);
        strcpy(curbp->b_name, fnlink);
    } else {
        if (s == (int) S_IFDIR) {
            strcpy(curbp->b_name, "? ");
            strcat(curbp->b_name, fnlink);
            strcat(curbp->b_name, "/");
        } else {
            strcpy(curbp->b_name, "?... ");
        }
        fname[0] = '\0';
        return (ABORT);
    }

    bp = curbp;                 /* Cheap.               */
    if ((s = clear_buffer(bp)) != TRUE) /* Might be old.        */
        return (s);
    bp->b_flag &= ~(BFINVS | BFCHG);

    /* let a user macro get hold of things */
    execute_key(&readhook, FALSE, 1);

    /* turn off ALL keyboard translation */
    TTkclose();

    if ((s = ffropen(fname)) == FIOERR) /* Hard file open.      */
        goto out;

    if (s == FIOFNF) {
        mlwrite(TEXT138);       /* "[New file]" */
        goto out;
    }

    /* read the file in */
    mlwrite(TEXT139);           /* "[Reading file]" */
    nline = 0;
    while ((s = ffgetline(&nbytes)) == FIOSUC) {
        if ((lp1 = alloc_line((size_t) nbytes)) == NULL) {
            s = FIOMEM;         /* Keep message on the  */
            break;              /* display.             */
        }
        lp2 = lback(curbp->b_linep);
        lp2->l_fp = lp1;
        lp1->l_fp = curbp->b_linep;
        lp1->l_bp = lp2;
        curbp->b_linep->l_bp = lp1;
        for (i = 0; i < nbytes; ++i)
            lputc(lp1, i, fline[i]);
        ++nline;
    }
    ffclose();                  /* Ignore errors.       */
    strcpy(mesg, "[");
    if (s == FIOERR) {
        strcat(mesg, TEXT141);  /* "I/O ERROR, " */
        curbp->b_flag |= BFTRUNC;
    }
    if (s == FIOMEM) {
        strcat(mesg, TEXT142);  /* "OUT OF MEMORY, " */
        curbp->b_flag |= BFTRUNC;
    }
    strcat(mesg, TEXT140);      /* "Read " */
    strcat(mesg, int2str(nline));
    strcat(mesg, TEXT143);      /* " line" */
    if (nline > 1)
        strcat(mesg, "s");
    strcat(mesg, "]");
    mlwrite(mesg);

  out:
    TTkopen();                  /* open the keyboard again */
    for (wp = wheadp; wp != NULL; wp = wp->w_wndp) {
        if (wp->w_bufp == curbp) {
            wp->w_linep = lforw(curbp->b_linep);
            wp->w_dotp = lforw(curbp->b_linep);
            wp->w_doto = 0;
            for (cmark = 0; cmark < NMARKS; cmark++) {
                wp->w_markp[cmark] = NULL;
                wp->w_marko[cmark] = 0;
            }
            wp->w_flag |= WFMODE | WFHARD;
        }
    }
    if (s == FIOERR || s == FIOFNF)     /* False if error.      */
        return (FALSE);
    return (TRUE);
}


/*++

Ask for a file name, and write the contents of the current buffer to that
file.  Update the remembered file name and clear the buffer changed flag.

++*/

int write_file(f, n)
int f, n;

{
    int s;
    char *fname;

    f = n = 0;                  /* splint */

    if ((fname = prompt4filename(TEXT144)) == NULL)     /* "Write file: " */
        return (FALSE);
    if ((s = writeout(fname, "w")) == TRUE) {
        strcpy(curbp->b_name, fname);
        curbp->b_flag &= ~BFCHG;
        /* Update mode lines.   */
        update_modeline();
    }
    return (s);
}



/*++

Interactive wrapper for writeout(fname, "a")

++*/
int append_file(f, n)
int f, n;

{
    int s;
    char *fname;

    f = n = 0;                  /* splint */

    if ((fname = prompt4filename(TEXT218)) == NULL)     /* "Append file: " */
        return (FALSE);
    if ((s = writeout(fname, "a")) == TRUE) {
        curbp->b_flag &= ~BFCHG;
        update_modeline();
    }
    return (s);
}



/*++

Save the contents of the current buffer in its associated file. Do
nothing if nothing has changed (this may be a bug, not a feature). Error
if there is no file name for the buffer.  Wrapper for writeout(fname,"w")

++*/
int save_file(f, n)
int f, n;

{
    int s;

    f = n = 0;                  /* splint */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    if ((curbp->b_flag & BFCHG) == 0)   /* Return, no changes.  */
        return (TRUE);

    /* complain about truncated files */
    if ((curbp->b_flag & BFTRUNC) != 0) {
        if (mlyesno(TEXT146) == FALSE) {        /* "Truncated file..write it out" */
            mlwrite(TEXT8);     /* "[Aborted]" */
            return (FALSE);
        }
    }

    /* complain about narrowed buffers */
    if ((curbp->b_flag & BFNAROW) != 0) {
        if (mlyesno(TEXT147) == FALSE) {        /*"Narrowed Buffer..write it out" */
            mlwrite(TEXT8);     /* "[Aborted]" */
            return (FALSE);
        }
    }

    if ((s = writeout(curbp->b_name, "w")) == TRUE) {
        curbp->b_flag &= ~BFCHG;
        /* Update mode lines.   */
        update_modeline();
    }
    return (s);
}



/*++

This function performs the details of file writing. It uses the file
management routines in the "fileio.c" package. The number of lines
written is displayed. Several errors are posible, and cause writeout
to return a FALSE result. When $ssave is TRUE, the buffer is written
out to a temporary file, and then the old file is unlinked and the
temporary renamed to the original name.  Before the file is written,
a user specifyable routine (in $writehook) can be run.

++*/
int writeout(fn, mode)
char *fn;                       /* name of file to write current buffer to */
char *mode;                     /* mode to open file (w = write a = append) */
{

    char *sp;                   /* temporary string pointer */
    char buf[NSTRING];          /* message buffer */
    char fnbak[LENPATH];        /* name of bak file */
    char tname[LENPATH];        /* temporary file name */
    int fflag;                  /* file flag (regular file?) */
    int nline;                  /* number of lines written */
    int sflag = FALSE;          /* backup switch */
    int status;                 /* return status */
    LINE *lp;                   /* line to scan while writing */
    struct stat st;


    execute_key(&writehook, FALSE, 1);

    if (fexist(fn) && *mode == 'w')
        sflag = TRUE;           /* backup~ */

    TTkclose();

    if (sflag) {
        sp = tname + (basename(fn) - fn) + 1;
        strcpy(tname, fn);
        do {
            *sp = 0;
            strcat(tname, int2str((int) time(NULL)));
        }
        while (fexist(tname));
        status = ffwopen(tname, "w");
    } else
        status = ffwopen(fn, mode);

    if (status != FIOSUC) {
        TTkopen();
        return (FALSE);
    }

    /* write the current buffer's lines to the open disk file */
    lp = lforw(curbp->b_linep);
    nline = 0;
    while (lp != curbp->b_linep) {
        if ((status = ffputline(&lp->l_text[0], lused(lp))) != FIOSUC)
            break;
        ++nline;
        lp = lforw(lp);
    }

    /* report on status of file write */
    *buf = 0;
    status |= ffclose();
    if (status == FIOSUC) {
        strcpy(buf, TEXT149);   /* "[Wrote " */
        strcat(buf, int2str(nline));
        strcat(buf, TEXT143);   /* " line" */
        if (nline > 1)
            strcat(buf, "s");

        if (sflag) {
            fflag = expandlink(fn, fnbak);
            if (fflag == 0) {
                strcpy(fn, fnbak);
                fflag = rename(tname, fn);
            } else if (fflag == (int) S_IFREG) {
                strcpy(fn, fnbak);
                strcat(fnbak, "~");
                unlink(fnbak);
                fflag = (rename(fn, fnbak) == 0);
                if (fflag) {
                    fflag = (rename(tname, fn) == 0);
                } else {
                    fflag = (unlink(fn) == 0);
                    if (fflag)
                        fflag = (rename(tname, fn) == 0);
                }
                if (fflag) {
                    stat(fnbak, &st);
                    chown(fn, st.st_uid, st.st_gid);
                    chmod(fn, st.st_mode);
                    strcpy(curbp->b_name, fn);
                } else
                    fflag = 0;
            }
            if (fflag) {
                strcpy(curbp->b_name, fn);
            } else {
                strcat(buf, TEXT150);   /* ", saved as " */
                strcat(buf, tname);
                strcpy(fn, tname);
                strcpy(curbp->b_name, tname);
                status = FIOSUC;
                /* but failed under original name! */
            }
        }
        strcat(buf, "]");
        mlwrite(buf);
    }

    /* reopen the keyboard, and return our status */
    TTkopen();
    return (status == FIOSUC);
}
